#ifndef __GeneradorArbre_H__
#define __GeneradorArbre_H__

#include "Ogre.h"
#include "OgreLogManager.h"
#include <iostream>
#include "NodeLlum.h"
#include "NodePuntual.h"
#include "NodeGrup.h"
#include "BoundingBox.h"

using namespace std;


class GeneradorArbre
{
public:
    /// Standard constructor
    GeneradorArbre()
    {
    }

	GeneradorArbre(std::deque<Light> llums)
	//Generem l'abre a partir de la llista "llums"
	//02-01-09 SEMBLA QUE VA BE. PROVAR AMB GENERACIO ALEATORIA DE REPRESENTANT!
    {
		llistaLlums = llums;
		NodeLlum * aux;
		BoundingBox * auxBB;

		//Creem les fulles de l'arbre, els nodes sense fills que representaran a una sola llum
		for (uint i=0;i<llums.size();i++){
			auxBB = new BoundingBox(llistaLlums[i].getPosition().x, llistaLlums[i].getPosition().y, llistaLlums[i].getPosition().z);
			aux = new NodePuntual(i,llistaLlums[i].getDiffuseColour(),llistaLlums[i].getSpecularColour(),llistaLlums[i].getDiffuseColour().b, auxBB);
			parellesLlums.push_back(aux);
		}
		
		//Combinem totes les fulles entre si per obtenir 
		parellesLlums = combinarParelles();

		//Entrem al metode vora
		while (parellesLlums.size() > 1){
			//Obtenim la primera parella de llums de la llista
			aux = getMillorParella();
			int noRep = aux->llumNoRepresentant();
			//Eliminar Parelles amb llum NO Representant
			eliminarNoRepresentant(noRep);
			//Substituir llum a les parelles que tinguin la llum Representant
			substituirRepresentant(aux);
		}

		arbre = parellesLlums[0];
    }

    ~GeneradorArbre()
    {

    }

	NodeLlum * getArrel(){
		return arbre;
	}

	Light getLlum(int pos){
		return llistaLlums[pos];
	}

	std::deque<Light> getLlistaLlums(){
		return llistaLlums;
	}

protected:
    std::deque<Light> llistaLlums;
	std::deque<NodeLlum*> parellesLlums;
	NodeLlum * arbre;

	Real distancia(int a, int b){
	/*Obtenim la distancia entre la llum 'a' i la llum 'b' del conjunt de parelles de llums 'parellesLlums'*/
		Light llumA = llistaLlums[parellesLlums[a]->llumRepresentant()];
		Light llumB = llistaLlums[parellesLlums[b]->llumRepresentant()];

		return llumA.getPosition().squaredDistance(llumB.getPosition());
	}

	NodeLlum * getMillorParella(){
	//Calculem totes les combinacions possibles de llums i retornem la millor
	//	std::deque<NodeLlum*> combinacionsLlums;
		
		//parellesLlums = combinarParelles();
		parellesLlums = ordenarParelles(parellesLlums);

		return parellesLlums[0];
	}

	std::deque<NodeLlum*> combinarParelles(){
	//Calculem totes les combinacions de parelles a partir de les de "parellesLlums"
		std::deque<NodeLlum*> combinacio;
		NodeLlum * aux;
		
		for (uint i=0;i<parellesLlums.size();i++){
			for (uint j=i+1;j<parellesLlums.size();j++){
				//creem la nova llum amb el seu fill esq i dret, el representant, la distancia i la intensitat mitjana
				aux = new NodeGrup(parellesLlums[i],
					parellesLlums[j],
					escollirRepresentant(parellesLlums[i]->llumRepresentant(),
					parellesLlums[j]->llumRepresentant()),
					distancia(i,j),
					parellesLlums[j]->getColorDifus()+parellesLlums[i]->getColorDifus(),
					parellesLlums[j]->getColorEspecular()+parellesLlums[i]->getColorEspecular(),
					parellesLlums[j]->getColorDifus().b + parellesLlums[i]->getColorDifus().b);
				aux->computeBoundingBox();
				combinacio.push_back(aux);
			}
		}

		return combinacio;

	}

	int escollirRepresentant(int i, int j){
	// Obtenim la llum representant de dues a l'atzar
		Real aleatori = (Real)Real(rand())/(Real)(RAND_MAX);

		if (aleatori > 0.5) return i;
		else return j;
	}

	std::deque<NodeLlum*> ordenarParelles(std::deque<NodeLlum*> combinacioLlums){
	//Utilitzem l'algoritme de la bombolla per ordenar les fulles
		std::deque<NodeLlum*> combinacio = combinacioLlums;

		for (uint i=1; i<combinacio.size();i++) {
			for(uint j=0;j<combinacio.size() - 1;j++){
				if (combinacio[j]->getDistancia() > combinacio[j+1]->getDistancia()) {
					NodeLlum * temp = combinacio[j]; 
					combinacio[j] = combinacio[j+1];
					combinacio[j+1] = temp;
				}
			}
		}
		return combinacio;
	}

	int getPosicio(int rep){
	//Obtenim la posicio a la llista de parelles d'una llum representant determinada
		uint i = 0;
		
		while (i < parellesLlums.size()){
			if (parellesLlums[i]->llumRepresentant() == rep)
				return i;
			else i++;
		}
		return -1;
	}

	void eliminarNoRepresentant(int noRep) {
	//Eliminar les parelles de la llista "parellesLlums" que tinguin la llum "noRep"
		std::deque<NodeLlum*> llistaAux;
		NodeLlum * parellaAux;
		for (uint i=0; i<parellesLlums.size();i++){
			parellaAux = parellesLlums[i];
			if (parellaAux->esNodeGrup() && parellaAux->fillDreta()->llumRepresentant() != noRep && parellaAux->fillEsq()->llumRepresentant() != noRep) {
				llistaAux.push_back(parellaAux);
			}
		}
		parellesLlums = llistaAux;
	}
    
	void substituirRepresentant(NodeLlum * llumRep) {
		NodeLlum * parellaAux;
		for (uint i=0; i<parellesLlums.size();i++){
			parellaAux = parellesLlums[i];
			if (parellaAux->esNodeGrup()) {
				if (parellaAux->fillDreta()->llumRepresentant() == llumRep->llumRepresentant()) {
					parellaAux->setLightDreta(llumRep);
					//Recalculem la intensitat per aquesta parella
					parellaAux->recalcularIntensitat();
				}
				if (parellaAux->fillEsq()->llumRepresentant() == llumRep->llumRepresentant()) {
					parellaAux->setLightEsq(llumRep);
					//Recalculem la intensitat per aquesta parella
					parellaAux->recalcularIntensitat();
				}
			}
		}
	}

};


#endif
